/************************************************************************************		
 * Wzorzec:			fabryka abstrakcyjna (abstract factory)								   *
 *																												*
 * Zastosowanie: 	Wzorzec ten umożliwia łatwe wprowadzanie zmian w zakresie klas    *
 *						używanych w kodzie, co umożliwia pisanie kodu np: na kilka 		   *
 *						platform jednocześnie, czy łatwiejsze zmienianie kodu wewnątrz    *
 *						klasy bibliotecznej.															   *
 *																												*	
 *	Opis:				Fabryka abstrakcyjna, zwykle tworzona jako interfejs, w języku	   *
 *						c++ jest implementowana za pomocą klasy abstrakcyjnej. Klasa ta   *
 *						ma wygląd typowej fabryki, jednak jako typy zwracane przez 		   *
 *						wirtualne metody fabrykujące ma zadeklarowane klasy abstrakcyjne 	*
 *						reprezentujące interfejsy obiektów zwracanych przez klasy 		   *
 *						pochodne. W kodzie programu implementuje się dla konkretnych 	   *
 *						zastosowań klasy pochodne od fabryki abstrakcyjnej wytwarzające   *
 *						obiekty odpowiednie dla danego systemu czy pliku 					   *
 *						konfiguracyjnego. Dzięki temu, że najczęściej fabryka jest 			*
 *						tworzona w programie tylko raz, przy zmianie zachowania obiektów	*
 *						wystarczy zmodyfikować tą linijkę w kodzie, w której następuje 	*
 *						przypisanie konkretnej fabryki.												*
 *																												*
 *	Przykładowy schemat (schemat opisuje inny przypadek niż kod, dla urozmaicenia) :	*
 *	Prezentowany jest przypadek programowania dla kilku systemów operacyjnych			*
 *																												*
 *						Odczyt schematu:																	*
 *						[Element na początku strzałki]												*
 *						(d) dziedziczy po																	*
 *						(w) zawiera wskaźnik/wskaźniki na											*
 *						(t) tworzy obiekty klasy														*
 *						[Element na który wskazuje strzałka]										*
 *					_______________				   ______________________						*
 *					| KlasaUsera  |-------------> |FabrykaAbstrakcyjna |						*
 *					|_____________|				(w)|____________________|						*
 *				   |_____________|					|____________________|						*
 *				   |_____________|					| A stworzObiekt()=0	|						*
 *							|								|____________________|						*
 *							|								  /|\(d)	/|\(d)  /|\(d)						*
 *							|				 ______________|		 |			|__________				*
 *							|		_______|__________ _________|________ ________|_________	*
 *							|		|FabrykaUnixowa  | |FabrykaDosowa	| |FabrykaWindows  |	*
 *	 -----------------/\----|________________| |________________| |________________|	*
 *	 |						|		|________________| |________________| |________________|	*
 *	 |						|		|A stworzObiekt()| |A stworzObiekt()| |A stworzObiekt()|	*
 *	 |					   |(w)	|________________| |________________| |________________|	*
 *  |		___________\|/_____									|						|				*
 *	 |		| A (klasa abstr)	|									|						|				*
 *	 |		|_________________|									|						|				*
 *	 |		|_________________|									|						|				*
 *	 |		| zareaguj()		|									|						|				*
 *	 |		|_________________|  								|						|				*
 *	 |		 /|\(d)	/|\(d)/|\(d)								|						|				*
*	 |(t)___|		 |		 |_________							|						|				*
 *	\|/__|______ ___|________ _____|______					|						|				*
 * | AUnix    | | ADos		| | AWindows |					|						|				*
 * |__________| |__________| |__________|	(t)			|						|				*
 *	|__________| |__________| |__________|<------------/\----------------|				*
 * |zareaguj()| |zareaguj()| |zareaguj()|					|										*
 *	|__________| |__________| |__________|					|										*
 *							/|\(t)									|										*
 *							 |_______________________________|										*
 *																												*
 ************************************************************************************/

/* Przykładowy kod. Przedstawia on rysowanie obiektów w zależności od
	szybkości procesora */
#include <iostream>//Biblioteka potrzebna w tym przykładzie
using namespace std;// dla wygody, żeby nie pisać wszędzie std

// Abstrakcyjny model obiektu graficznego
class ObiektGraficzny_Abstrakcyjny {
	protected:	
		float x;// położenie na ekranie w osi x
		float y;// położenie na ekranie w osi y
	public:
			virtual void rysujSie()=0;// metoda do rysowania obiekt
			// poniższe metody są wirtualne, gdyż klasy pochodne nie muszą koniecznie w ten sposób przechowywać
			// położenia obiektu, jednak mają już ciała, gdyż to będzie najczęstszy przypadek

			// metoda do ustalania położenia obiektu
			virtual void ustawPozycje(float x, float y){
				this->x=x;
				this->y=y;			
			}
			// metody do pobierania położenia obiektu
			virtual float podajPozX() {return x;}
			virtual float podajPozY() {return y;}
			ObiektGraficzny_Abstrakcyjny (float newX, float newY) : x(newX), y(newY){}
			virtual ~ObiektGraficzny_Abstrakcyjny() {}//wirtualny destruktor, bo są wirtualne funkcje
};

// Abstrakcyjna fabryka obiektów, służąca za interfejs
class FabrykaObiektowGraficznych_Abstrakcyjna {
	public:
			// metoda do tworzenia instancji obiektów
			virtual ObiektGraficzny_Abstrakcyjny * utworzObiekt(float x, float y)=0;
			virtual ~FabrykaObiektowGraficznych_Abstrakcyjna(){}//wirtualny destruktor, bo sa wirtualne funkcje
}; 

// Konkretny model obiektu graficznego, dla wolniejszych komputerów
class ObiektGraficzny_Prosty : public ObiektGraficzny_Abstrakcyjny {
	public:
			// funkcja rysujSie jest mocno okrojona, gdyż to tylko przykład zastosowania
			void rysujSie(){
				cout<<"malo wymagajacy model jest rysowany na pozycji: "<<x<<", "<<y<<endl;
			}
			ObiektGraficzny_Prosty (float x, float y):ObiektGraficzny_Abstrakcyjny(x,y){}
};

// Konkretny model obiektu graficznego, dla szybszych komputerów
class ObiektGraficzny_Trudny : public ObiektGraficzny_Abstrakcyjny {
	public:
			// funkcja rysujSie jest mocno okrojona, gdyż to tylko przykład zastosowania
			void rysujSie(){
				cout<<"BARDZO WYMAGAJACY MODEL JEST RYSOWANY NA POZYCJI: "<<x<<", "<<y<<endl;
			}
			ObiektGraficzny_Trudny (float x, float y):ObiektGraficzny_Abstrakcyjny(x,y){}
};

//Konkretna fabryka obiektów graficznych, dla wolnych komputerów
class FabrykaObiektowGraficznych_Prostych : public FabrykaObiektowGraficznych_Abstrakcyjna {
	public:
		// skonkretyzowanie metody
		ObiektGraficzny_Abstrakcyjny * utworzObiekt(float x, float y) {
				return new ObiektGraficzny_Prosty(x,y);		
		}
};

//Konkretna fabryka obiektów graficznych, dla szybkich komputerów
class FabrykaObiektowGraficznych_Trudnych : public FabrykaObiektowGraficznych_Abstrakcyjna {
	public:
		// skonkretyzowanie metody
		ObiektGraficzny_Abstrakcyjny * utworzObiekt(float x, float y) {
				return new ObiektGraficzny_Trudny(x,y);		
		}
};

// Klasa korzystająca z fabryk
class KlasaUzytkowa {
	private:
		static const int ILE_OBIEKTOW=5;//ile obiektow graficznych naraz na ekranie maksymalnie
		// Wskaznik na fabryke obiektow, tu bedzie podany wskaznik do konkretnej fabryki 
	 	FabrykaObiektowGraficznych_Abstrakcyjna * fabryka;
		ObiektGraficzny_Abstrakcyjny * pulaObiektow[ILE_OBIEKTOW]; // Tablica obiektow do rysowania
	public:
		// dla zapełniania tablicy obiektów, zmienna nowa partia mówi czy zmieniać obiekty,
		// czy też tworzyć nowe od poczatku
		void pobierzNowe(bool nowaPartia){
			for (int licznik=0; licznik<5; ++licznik){
				ObiektGraficzny_Abstrakcyjny * temp=pulaObiektow[licznik];			
				if (temp!=NULL){// odtworzenie obiektow w nowych warunkach
						pulaObiektow[licznik]=fabryka->utworzObiekt(temp->podajPozX(), temp->podajPozY());	
						delete temp; 			
				}
				else if (nowaPartia){
					pulaObiektow[licznik]=fabryka->utworzObiekt(licznik, ILE_OBIEKTOW - licznik);	
				}				
			}		
		}
		// dla ustawienia źródła obiektów
		void ustawFabryke(FabrykaObiektowGraficznych_Abstrakcyjna * nowaFabryka) {
				delete fabryka;// usunięcie starej fabryki				
				fabryka=nowaFabryka;	// ustawienie nowej fabryki
				pobierzNowe(false);//żeby od razu zmienić typ przechowywanych obiektów	
		} 
		// dla rysowania obiektów
		void rysujObiekty(){
			for (int licznik=0; licznik<ILE_OBIEKTOW; ++licznik){
				if (pulaObiektow[licznik]!=NULL){
						pulaObiektow[licznik]->rysujSie();				
				}			
			}		
		}
		// konstruktor
		KlasaUzytkowa (FabrykaObiektowGraficznych_Abstrakcyjna *nowaFabryka){
				fabryka=nowaFabryka;		
		}
		// destruktor
		~KlasaUzytkowa(){
			for (int licznik=0; licznik<5; ++licznik){
				delete pulaObiektow[licznik];// zwolnienie puli obiektów		
			}			
			delete fabryka;// zwolnienie fabryki
		}		
};

//Program testowy
int main(){
		cout<<"..:: Uruchamianie programu w trybie wysokiej jakosci grafiki ::.."<<endl;
		// Na początek ustawiany bardziej wymagający typ obiektu graficznego
		KlasaUzytkowa * testowa=new KlasaUzytkowa(new FabrykaObiektowGraficznych_Trudnych);
		testowa->pobierzNowe(true);// Ustawianie przykładowych obiektów
		cout<<"..:: Rysowanie przykladowych obiektow ::.."<<endl;
		testowa->rysujObiekty();
		//{...} Tutaj okazało się, że niestety maszyna jest za wolna
		cout<<"..:: Niestety maszyna jest za wolna, przelaczanie w tryb niskiej jakosci::.."<<endl;
		testowa->ustawFabryke(new FabrykaObiektowGraficznych_Prostych());
		cout<<"..:: Rysowanie przykladowych obiektow ::.."<<endl;
		testowa->rysujObiekty();
		cout<<"..:: Koniec programu ::.."<<endl;
}

